/*
	$Id: Downloader.java,v 1.2 2003/06/15 06:19:58 admin Exp $
	-----
	
    SSTP Bottle Client for MacOS X, using protocol SLPP2.0
    Copyright (C) 2001-2002 phonohawk.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package gui.logdownload;
import gui.logdownload.*;
import core.*;
import log.*;
import gui.logbrowser.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.util.regex.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import com.apple.cocoa.application.*;
import com.apple.cocoa.foundation.*;

public class Downloader extends Thread
{
	LogDownloadController controller;
	int year,month,day;
	
	public Downloader(LogDownloadController controller,int year,int month,int day)
	{
		this.controller = controller;
		this.year = year;
		this.month = month;
		this.day = day;
	}
	
	public void run()
	{
		int pool = NSAutoreleasePool.push();
		
		controller.setEnabledGetButton(false);
		try {
			_run();
		} catch (Exception e) {
			// ByteArrayOutputStreamにスタックトレースを吐かせる。
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			e.printStackTrace(new PrintWriter(baos));
			controller.setMessage(baos.toString());
			
			// 標準エラーにも吐いとく。
			e.printStackTrace();
		}
		controller.setEnabledGetButton(true);
		controller.setProgressState(0);
		
		NSAutoreleasePool.pop(pool);
	}
	
	protected void _run() throws Exception
	{
		// 瓶を持ってくる
		Vector bottles = parseXmlDataStream(getXmlDataStream());
		
		/*
			マージは次のように行なう。
			
			XMLの取得中に瓶が増えたとしても、XMLログの最終瓶のMID時刻以前の瓶は全てログに入っていた事が保証される。
			従って、まずはローカルログの終わりから一つずつMID時刻を見ていって、XMLログの最終MID時刻より後のものがあれば
			それらを全てログに追加する。
			
			次に、XMLの取得中に投票/同意数が増加した可能性があるため、ローカルログの方が信用出来る。
			従って、ログの全ての瓶の投票/同意数を、ローカルログからコピーする。
		*/
		
		// ローカルログを取得してロック。
		controller.setMessage("Getting local saved log...");
		LogData local = LogDataServer.sharedLogServer().get(year,month,day);
		
		Vector local_vec = local.getArray(LogData.LOG_ALL);
		synchronized (local_vec) {
			controller.setMessage("Merging...");
			
			// 最終瓶のMIDをコピーしておく。
			String xml_lastbottle_mid = (String)((Hashtable)bottles.elementAt(bottles.size()-1)).get("mid");
			
			// 追加すべき瓶を集める。
			Vector bottles_to_append = new Vector();
			for (int i = local.size()-1; i >= 0; i--) {
				Hashtable bottle = (Hashtable)local.elementAt(i);
				
				// この瓶はxml最終瓶より後のものか？
				if (MidAnalyzer.compare((String)bottle.get("mid"),xml_lastbottle_mid) == 1) {
					bottles_to_append.add(0,bottle);
				}
				else {
					// これより前のローカルボトルはチェックする必要が無い。
					break;
				}
			}
			
			// 同意/投票数をローカルからコピー。
			// 量が量なのでコピー先(つまりbottles)は一旦インディクスを張ってからコピーを行なう。
			Map bottles_map = new Hashtable(bottles.size());
			for (Enumeration e = bottles.elements(); e.hasMoreElements(); ) {
				Hashtable bottle = (Hashtable)e.nextElement();
				
				bottles_map.put((String)bottle.get("mid"),bottle);
			}
			for (Enumeration e = local.elements(); e.hasMoreElements(); ) {
				Hashtable src = (Hashtable)e.nextElement();
				Hashtable dest = (Hashtable)bottles_map.get((String)src.get("mid"));
				if (dest != null) {
					if (src.get("vote") != null) {
						dest.put("vote",src.get("vote"));
					}
					if (src.get("agree") != null) {
						dest.put("agree",src.get("agree"));
					}
				}
			}
			
			// この内容でローカルログを置き換えて、グループを再構成した後にdirtyフラグを立てる。
			local_vec.clear();
			local_vec.addAll(bottles);
			local_vec.addAll(bottles_to_append);
			local.reconstructGroups();
			local.beDirty(LogData.LOG_ALL);
		}
		
		// commitした後、ログブラウザで現在表示中のログが今回のものだったら、リロードさせる。
		controller.setMessage("Saving...");
		controller.setProgressState(-1);
		local.commit();
		
		LogBrowserController log_browser = LogBrowserController.sharedLogBrowser();
		log_browser.reloadTableIfShowing(year,month,day);
		
		controller.setMessage("Done.");
		controller.setProgressState(0);
	}
	
	protected InputStream getXmlDataStream() throws Exception
	{
		// http://bottle.mikage.to/viewlog2.cgi?days=731669&start=0&channel=&style=xml
		// daysの部分は-1年1月3日からの経過日数となっている。
		// 何故そんな仕様なのか知らないがとにかくそうなっているので…
		Calendar cal = Calendar.getInstance();
		cal.set(Calendar.YEAR,year);
		cal.set(Calendar.MONTH,month-1);
		cal.set(Calendar.DAY_OF_MONTH,day);
		long time_to_get = cal.getTime().getTime();
		
		cal.set(Calendar.YEAR,2001);
		cal.set(Calendar.MONTH,0);
		cal.set(Calendar.DAY_OF_MONTH,8);
		long time_origin = cal.getTime().getTime();
		
		// ダウンロードするファイルのパスを作る。
		long days_after_origin = ((time_to_get - time_origin) / (1000*60*60*24)) + 730858;
		String path = "/viewlog2.cgi?days="+days_after_origin+"&start=0&channel=&style=xml";
		
		
		String app_name = NSBundle.mainBundle().localizedStringForKey("CFBundleShortVersionString","BotCli","InfoPlist");
        
		NSUserDefaults defaults = NSUserDefaults.standardUserDefaults();
		String botservername = (String)defaults.objectForKey("bottleserver.servername");
		if (botservername == null) botservername = "bottle.mikage.to";
		int httpport = defaults.integerForKey("bottleserver.httpport");
		if (httpport == 0) httpport = 80;
					
		boolean use_proxy = defaults.integerForKey("proxy.http.use") == 1;
		String http_proxy_host = (String)defaults.objectForKey("proxy.http.host");
		if (http_proxy_host == null || http_proxy_host.length() == 0) use_proxy = false;
		int http_proxy_port = defaults.integerForKey("proxy.http.port");
		if (defaults.objectForKey("proxy.http.port") == null) use_proxy = false;
		
		controller.setProgressState(-1);
		
		Socket s;
		if (use_proxy) {
			controller.setMessage("Connecting to "+http_proxy_host+":"+http_proxy_port+" ...");
			s = new Socket(http_proxy_host,http_proxy_port);
		}
		else {
			controller.setMessage("Connecting to "+botservername+":"+httpport+" ...");
			s = new Socket(botservername,httpport);
		}
		
		controller.setMessage("Sending request for the xml log...");
		if (use_proxy) {
			sockwrite(s,"GET http://"+botservername+":"+httpport+path+" HTTP/1.0\r\n");
		}
		else {
			sockwrite(s,"GET "+path+" HTTP/1.0\r\n");
		}
		sockwrite(s,"Host: "+botservername+":"+httpport+"\r\n");
		sockwrite(s,"User-Agent: "+app_name+"\r\n");
		sockwrite(s,"\r\n");
		
		// ステータス行を読む。
		String status = sockreadline(s);
		if (!Pattern.matches("^HTTP/.+? 200 OK$",status)) {
			// 200 OKでなかったら終了…
			controller.setMessage("Server replied: "+status);
			throw new Exception("Server replied: "+status);
		}
		
		// 空の行が来るまではヘッダ。
		Map header = new Hashtable();
		while (true) {
			String line = sockreadline(s).trim();
			if (line.length() == 0) {
				// 空の行だ。ヘッダ終わり。
				break;
			}
			else {
				Matcher m = Pattern.compile("^(.+?)\\s*:\\s*(.+?)$").matcher(line);
				if (m.matches()) {
					header.put(m.group(1),m.group(2));
				}
			}
		}
		
		// 興味があるのはContent-Lengthのみ。
		// 無ければ無いで構わない。
		int content_length;
		try {
			content_length = Integer.parseInt((String)header.get("Content-Length"));
			controller.setProgressState(0);
		}
		catch (Exception e) {
			content_length = -1;
			controller.setProgressState(-1);
		}
		
		// 中身の受信を開始
		byte[] buf = new byte[1024 * 100]; // 100Kb buffer
		ByteArrayOutputStream baos = new ByteArrayOutputStream(content_length == -1 ? 1024 * 500 : content_length);
		long received_bytes = 0;
		// 0.5秒に一度だけ表示を更新する。
		// 落　ち　な　い　よ　う　に
		long last_indicator_updated_time = 0;
		while (true) {
			int read_bytes = s.getInputStream().read(buf);
			if (read_bytes == -1) {
				// EOF
				break;
			}
			else {
				baos.write(buf,0,read_bytes);
				received_bytes += read_bytes;
				
				if (last_indicator_updated_time + 500 < System.currentTimeMillis()) {
					last_indicator_updated_time = System.currentTimeMillis();
					
					if (content_length == -1) {
						controller.setMessage("Downloading XML log...\n"+
											"Received: "+received_bytes+" bytes");
					}
					else {
						controller.setMessage("Downloading XML log...\n"+
											"Received: "+received_bytes+" bytes\n"+
											"Exists: "+content_length+" bytes\n"+
											"Remaining: "+(content_length - received_bytes)+" bytes");
						controller.setProgressState((double)received_bytes / (double)content_length);
					}
				}
			}
		}
		
		// 受信完了。
		s.close();
		return new ByteArrayInputStream(baos.toByteArray());
	}
	
	public void sockwrite(Socket s,String str) throws Exception
	{
		OutputStream os = s.getOutputStream();
		os.write(str.getBytes());
		os.flush();
	}
	
	public String sockreadline(Socket s) throws Exception
	{
		// CRLFが来るまで読み続け、来たらそれまでに受け取った行を返す。
		// 返される文字列にCRLFそのものは含まれない。
		InputStream is = s.getInputStream();
		StringBuffer buf = new StringBuffer();
		boolean CR_received = false;
		while (true) {
			int read = is.read();
			if (read == -1) {
				// EOF
				break;
			}
			else {
				char c = (char)read;
				if (c == '\r') {
					CR_received = true;
				}
				else if (c == '\n') {
					if (CR_received) {
						break;
					}
					else {
						CR_received = false;
						buf.append(c);
					}
				}
				else {
					if (CR_received) {
						buf.append('\r');
					}
					CR_received = false;
					buf.append(c);
				}
			}
		}
		return buf.toString();
	}
	
	protected Vector parseXmlDataStream(InputStream is) throws Exception
	{
		// isがavailable()が読み込み可能な全バイト数を返すようなInputStreamであった場合のみ
		// プログレスバーに正しい値を設定します。例えばByteArrayInputStreamはこの条件を満たします。
		SAXParserFactory factory = SAXParserFactory.newInstance();
		SAXParser parser = factory.newSAXParser();
		
		controller.setMessage("Parsing XML log...\n"+
							  "Size: "+is.available());
		controller.setProgressState(0);
		
		Vector bottles = new Vector();
		parser.parse(is,new Handler_bottlelog(controller,is,bottles));
		return bottles;
	}
	
	protected class Handler_bottlelog extends DefaultHandler
	{
		LogDownloadController controller;
		InputStream is;
		int size;
		Vector bottles;
		Hashtable current_bottle;
		String current_tag;
		StringBuffer current_text;
		
		public Handler_bottlelog(LogDownloadController controller,InputStream is,Vector bottles) throws Exception
		{
			this.controller = controller;
			this.is = is;
			this.size = is.available();
			this.bottles = bottles;
			this.current_bottle = null;
			this.current_text = new StringBuffer();
		}
		
		public void startElement(String uri,String localName,String qName,Attributes attributes)
		{
			if (qName.equals("message")) {
				// 新たな瓶を準備。
				current_bottle = new Hashtable();
				
				// midを保存すると同時に、それを解析してdateとtimeを作る。
				String mid = attributes.getValue("mid");
				current_bottle.put("mid",mid);
				
				Calendar cal = MidAnalyzer.analyzeTimeStampFromMid(mid);
				
				String date = pad(cal.get(Calendar.MONTH)+1,2)+pad(cal.get(Calendar.DAY_OF_MONTH),2)+pad(cal.get(Calendar.YEAR),4);
				current_bottle.put("date",date);
				
				String time = pad(cal.get(Calendar.HOUR_OF_DAY),2)+pad(cal.get(Calendar.MINUTE),2)+pad(cal.get(Calendar.SECOND),2);
				current_bottle.put("time",time);
			}
			else {
				current_tag = qName;
				current_text.setLength(0);
			}
		}
		
		public void characters(char[] ch,int offset,int length)
		{
			String content = new String(ch,offset,length).trim();
			current_text.append(content);
		}
		
		public void endElement(String uri,String localName,String qName)
		{
			if (qName.equals("message")) {
				// この瓶終わり。
				bottles.add(current_bottle);
				
				try {
					controller.setMessage("Parsing XML log...\n"+
										"XML Size: "+size+" bytes\n"+
										"Parsed: "+(size - is.available())+" bytes\n"+
										"Remaining: "+is.available()+" bytes\n"+
										"Parsed Bottles: "+bottles.size()+" bottles");
					controller.setProgressState((double)(size - is.available()) / (double)size);
				} catch (Exception e) {}
			}
			else {
				String content = current_text.toString();
				
				if (qName.equals("channel")) {
					current_bottle.put("channel",content);
				}
				else if (current_tag.equals("script")) {
					current_bottle.put("script",content);
				}
				else if (current_tag.equals("votes")) {
					current_bottle.put("vote",content);
				}
				else if (current_tag.equals("agrees")) {
					current_bottle.put("agree",content);
				}
				else if (current_tag.equals("ghost")) {
					current_bottle.put("ifghost",content);
				}
			}
		}
	}
	
	public static String pad(int n,int chars)
	{
		StringBuffer buf = new StringBuffer(chars);
		String str_n = Integer.toString(n);
		
		int n_pad = chars - str_n.length();
		for (int i = 0; i < n_pad; i++) {
			buf.append('0');
		}
		buf.append(str_n);
		
		return buf.toString();
	}
}
