/*
	$Id: MainWindowController.java,v 1.1.1.1 2003/04/08 14:55:46 admin Exp $
	-----
	
    SSTP Bottle Client for MacOS X, using protocol SLPP2.0
    Copyright (C) 2001-2002 phonohawk.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package gui.main;
import gui.logbrowser.*;
import gui.logdownload.*;
import gui.main.*;
import gui.yenurl.*;
import core.*;
import com.apple.cocoa.foundation.*;
import com.apple.cocoa.application.*;
import java.io.*;
import java.util.*;

public class MainWindowController extends NSWindowController
{
    private static MainWindowController _sharedInstance = null;
	
	public static final int LIMIT_OF_SCRIPT_LENGTH = 1900;
    
    ChannelInfoViewDataSource chInfoDataSource;
    MainWindowGhostComboDataSource scriptsend_combo_ghost_datasource;
    
    // outlets
    NSTableView channelListView;
    NSTextView scriptView;
    
    NSButton serverstats_btn_toggleconnect;
    NSTextField serverstats_channels;
    NSTextField serverstats_users;
    NSButton serverstats_connected;
    NSButton serverstats_receive;
    NSButton serverstats_send;

    NSTextField sstpqueue_inqueue;
    
    NSButton scriptsend_localtest;
    NSButton scriptsend_clear; // \{^ResetɕύX܂B
    NSButton scriptsend_broadcast;
    NSComboBox scriptsend_combo_ghost;
	
	NSPopUpButton tags_s;
	
	NSTextField lengthofscript_current;
	NSTextField lengthofscript_limit;
    
    NSButton channel_login; // Join
    NSButton channel_logout; // Leave
    NSButton channel_selectall;
    NSTextField channel_registered;
    
    public static MainWindowController sharedInstance()
    {
        if (_sharedInstance == null)
        {
            _sharedInstance = new MainWindowController();
        }
        return _sharedInstance;
    }
    
    private MainWindowController()
    {
        super("MainWindow");
        loadWindow();
	
		window().setFrameUsingName("mainwindow_frame");
        window().setFrameAutosaveName("mainwindow_frame");
		scriptsend_clear(null);
        tableViewSelectionDidChange(null);
        textDidChange(null);
        
        chInfoDataSource = new ChannelInfoViewDataSource();
        channelListView.setDataSource(chInfoDataSource);
	
		scriptsend_combo_ghost_datasource = new MainWindowGhostComboDataSource();
		scriptsend_combo_ghost.setDataSource(scriptsend_combo_ghost_datasource);
		scriptsend_combo_ghost.selectItemAtIndex(0); // Default
		scriptsend_combo_ghost.setNumberOfVisibleItems(10);
		broadcasting_ghost_changed(null);
        
        String app_name = NSBundle.mainBundle().localizedStringForKey("CFBundleShortVersionString","SSTP Bottle Client for MacOS X","InfoPlist");
        window().setTitle(app_name);
	
		NSNotificationCenter.defaultCenter().addObserver(
							this,
							new NSSelector("notification_reload",new Class[] {NSNotification.class}),
							"BCGhostSpecModified",
							null);
		
		// ~bgݒ
		lengthofscript_limit.setStringValue(LIMIT_OF_SCRIPT_LENGTH+" bytes");
		
		// XNvg̓r[first responderɂB
		window().makeFirstResponder(scriptView);
    }
    
    public void setConnectedState(boolean stat)
    {
        serverstats_connected.setState(stat ? NSCell.OnState : NSCell.OffState);
		serverstats_btn_toggleconnect.setTitle(stat ? "Disconnect" : "Connect");
		
		LogBrowserController.sharedLogBrowser().buttons_validate();
    }
    
    public void setReceiveState(boolean stat)
    {
        serverstats_receive.setState(stat ? NSCell.OnState : NSCell.OffState);
    }
    
    public void setSendState(boolean stat)
    {
        serverstats_send.setState(stat ? NSCell.OnState : NSCell.OffState);
    }
    
    public void setUsersCount(int c)
    {
        // -1nꂽꍇ͋󔒂ɂ܂B
        if (c == -1) {
            serverstats_users.setStringValue("");
        }
        else {
            serverstats_users.setIntValue(c);
        }
    }
    
    public void setChannelsCount(int c)
    {
        // -1nꂽꍇ͋󔒂ɂ܂B
        if (c == -1) {
            serverstats_channels.setStringValue("");
        }
        else {
            serverstats_channels.setIntValue(c);
        }
    }
    
    public void setChannelRegisteredCount(int c)
    {
        // -1nꂽꍇ͋󔒂ɂ܂B
        if (c == -1) {
            channel_registered.setStringValue("");
        }
        else {
            channel_registered.setIntValue(c);
        }
    }
    
    public void setBottlesInQueueCount(int c)
    {
        // -1nꂽꍇ͋󔒂ɂ܂B
        if (c == -1) {
            sstpqueue_inqueue.setStringValue("");
        }
        else {
            sstpqueue_inqueue.setIntValue(c);
        }
    }
    
    public void reloadChListView()
    {
        channelListView.reloadData();
        tableViewSelectionDidChange(null);
    }
    
    // notification handlers
    public void notification_reload(NSNotification notification)
    {
		scriptsend_combo_ghost_datasource.reloadFromDefaults();
		scriptsend_combo_ghost.reloadData();
    }
    
    // Actions
    public void channel_login(Object sender)
    {
        // channelListViewőIĂ`lo^܂B
        // ۂ̓o^ServerConnectionSLPP2.0 CGIɃR}h|Xg邱Ƃōs܂B
        if (!Foundation.sharedFoundation().slpp_connected()) return;
        ServerConnection sc = Foundation.sharedFoundation().getSlppConnection();
        Vector chList = sc.getChannelList();
        
        Vector channels = new Vector(); // StringzBg̓`lB
        int n_ch = channelListView.numberOfRows();
        for (int i = 0;i < n_ch;i++)
        {
            if (channelListView.isRowSelected(i))
            {
                ChannelInfo info = (ChannelInfo)chList.elementAt(i);
                channels.add(info.name());
            }
        }
        
        sc.loginChannels(channels);
    }
    
    public void channel_logout(Object sender)
    {
        // channelListViewőIĂ`l̓o^܂B
        // ۂ̓o^ServerConnectionSLPP2.0 CGIɃR}h|Xg邱Ƃōs܂B
        if (!Foundation.sharedFoundation().slpp_connected()) return;
        ServerConnection sc = Foundation.sharedFoundation().getSlppConnection();
        Vector chList = sc.getChannelList();
        
        Vector channels = new Vector(); // StringzBg̓`lB
        int n_ch = channelListView.numberOfRows();
        for (int i = 0;i < n_ch;i++)
        {
            if (channelListView.isRowSelected(i))
            {
                ChannelInfo info = (ChannelInfo)chList.elementAt(i);
                channels.add(info.name());
            }
        }
        
        sc.logoutChannels(channels);
    }
    
    public void channel_selectall(Object sender)
    {
        channelListView.selectAll(null);
    }
    
    public void queue_sleep(NSButton sender)
    {
        SSTPQueue queueManager = Foundation.sharedFoundation().getQueue();
        queueManager.setSleeping(sender.state() == NSCell.OnState);
    }
	
	public void queue_clear(NSButton sender)
	{
		SSTPQueue queueManager = Foundation.sharedFoundation().getQueue();
		queueManager.clear();
	}
	
	public void broadcasting_ghost_changed(Object sender)
	{
		// sender͖B
		String ifghost = scriptsend_combo_ghost.stringValue();
		SurfaceDescription.GhostInfo info = SurfaceDescription.sharedDescription().get(ifghost);
		if (info == null) {
			info = SurfaceDescription.sharedDescription().get(SurfaceDescription.DEFAULT);
		}
		
		NSMenu menu = tags_s.menu();
		menu.setAutoenablesItems(false);
		// ܂͓ڈȍ~Sč폜
		final int n_items = menu.numberOfItems();
		for (int i = 1; i < n_items; i++) {
			menu.removeItemAtIndex(1);
		}
		// sakura
		NSMenuItem sakura_label = new NSMenuItem(info.getSakuraName(),null,"");
		sakura_label.setEnabled(false);
		menu.addItem(sakura_label);
		for (Enumeration e = info.getSakuraSurfaces(); e.hasMoreElements(); ) {
			Object[] surface = (Object[])e.nextElement();
			
			NSMenuItem item = new NSMenuItem((String)surface[1],null,"");
			item.setTag(((Integer)surface[0]).intValue());
			item.setEnabled(true);
			
			menu.addItem(item);
		}
		// d؂
		menu.addItem((NSMenuItem)NSMenuItem.protocolSeparatorItem());
		// kero
		NSMenuItem kero_label = new NSMenuItem(info.getKeroName(),null,"");
		kero_label.setEnabled(false);
		menu.addItem(kero_label);
		for (Enumeration e = info.getKeroSurfaces(); e.hasMoreElements(); ) {
			Object[] surface = (Object[])e.nextElement();
			
			NSMenuItem item = new NSMenuItem((String)surface[1],null,"");
			item.setTag(((Integer)surface[0]).intValue());
			item.setEnabled(true);
			
			menu.addItem(item);
		}
	}
	
	public static String getInitialScriptFormat()
	{
		NSUserDefaults defaults = NSUserDefaults.standardUserDefaults();
		String format = (String)defaults.objectForKey("iniscript.script");
		if (format == null) {
			return "\u00a5t\u00a5u\u00a5s[10]\u00a5h\u00a5s[0]|\u00a5e";
		}
		else {
			return format;
		}
	}
	
    public void scriptsend_clear(Object sender)
    {
		String format = getInitialScriptFormat();
		
		final int firstVerticalLinePos = format.indexOf('|');
		if (firstVerticalLinePos != -1)
		{
			// cSď
			StringBuffer sb = new StringBuffer(format);
			while (true)
			{
				int vertLinePos = sb.toString().indexOf('|');
				if (vertLinePos == -1) break;
				
				sb.deleteCharAt(vertLinePos);
			}
			
			scriptView.setString(sb.toString());
			scriptView.setSelectedRange(new NSRange(firstVerticalLinePos,0));
		}
		else
		{
			scriptView.setString(format);
		}
		textDidChange(null);
    }
    
    public void scriptsend_localtest(Object sender)
    {
        String script = validateScript(scriptView.string());
        String app_name = NSBundle.mainBundle().localizedStringForKey("CFBundleShortVersionString","BotCli","InfoPlist");
        
		String ghost = null;
		if (channelListView.numberOfSelectedRows() == 1)
        {
            ServerConnection sc = Foundation.sharedFoundation().getSlppConnection();
            Vector chList = sc.getChannelList();
            int ch_id = channelListView.selectedRow();
            
            ghost = ((ChannelInfo)chList.elementAt(ch_id)).ghost();
        }
	
		String ifghost = scriptsend_combo_ghost.stringValue();
		if (ifghost.equals(MainWindowGhostComboDataSource.DEFAULT_ITEM))
		{
			ifghost = null; // # Default #
		}
		
		String target = null;
		if (ifghost != null) target = ifghost;
		else if (ghost != null) target = ghost;
	
        Vector req = new Vector();
        req.add("SEND SSTP/1.4");
        req.add("Sender: "+app_name+" local-test;"+(target == null ? " " : " [Target:"+target+"]"));
        if (target != null) req.add("IfGhost: "+target);
        req.add("Script: "+ScriptAnalyzer.expandYenURL(script));
        req.add("Charset: SJIS");
        
        //Foundation.sharedFoundation().getQueue().push(req);
		Foundation.sharedFoundation().sendToLocalSstpd(req);
    }

    public void scriptsend_broadcast(Object sender)
    {
        if (channelListView.numberOfSelectedRows() != 1) return;
        
        ServerConnection sc = Foundation.sharedFoundation().getSlppConnection();
        Vector chList = sc.getChannelList();
        
        int ch_id = channelListView.selectedRow();
        
        String script = validateScript(scriptView.string());
        String chname = ((ChannelInfo)chList.elementAt(ch_id)).name();
		String ifghost = scriptsend_combo_ghost.stringValue();
		if (ifghost.equals(MainWindowGhostComboDataSource.DEFAULT_ITEM))
		{
			ifghost = null; // # Default #
		}
        
        sc.broadcast(script,chname,ifghost);
    }
    
    private String validateScript(String param)
    {
        // snɒuAeŏIĂȂ炻tB
        StringBuffer sb = new StringBuffer(param);
        int return_pos;
        while ((return_pos = sb.toString().indexOf('\n')) != -1)
        {
            sb.replace(return_pos,return_pos + 1,"\u00a5n");
        }
        if (!sb.toString().endsWith("\u00a5e")) sb.append("\u00a5e");
        
        return sb.toString();
    }

    public void serverstats_toggleconnect(Object sender)
    {
        if (Foundation.sharedFoundation().slpp_connected())
        {
            //Foundation.sharedFoundation().slpp_disconnect(); // ؒf
			new SlppTogglerThread(false).start();
        }
        else
        {
            //Foundation.sharedFoundation().slpp_connect(); // ڑ
			new SlppTogglerThread(true).start();
        }
    }
	private class SlppTogglerThread extends Thread
	{
		// ꂽ{^
		boolean state_to_set;
		public SlppTogglerThread(boolean state_to_set) {
			this.state_to_set = state_to_set;
		}
		public void run() {
			int pool = NSAutoreleasePool.push();
			if (state_to_set) {
				Foundation.sharedFoundation().slpp_connect(); // ڑ
			}
			else {
				Foundation.sharedFoundation().slpp_disconnect(); // ؒf
			}
			NSAutoreleasePool.pop(pool);
		}
	}

    public void tags_s(NSPopUpButton sender)
    {
        scriptView.insertText("\u00a5s["+sender.selectedItem().tag()+']');
    }

    public void tags_simple(NSButton sender)
    {
        scriptView.insertText(sender.title());
    }

    public void tags_url(Object sender)
    {
		YenURLAssistantController.sharedController().showWindow(null);
    }

    public void tags_w(NSPopUpButton sender)
    {
        scriptView.insertText("\u00a5w"+sender.selectedItem().tag());
    }
    
    public void tags_percent(NSPopUpButton sender)
    {
        String title = sender.selectedItem().title();
        int sp_pos;
        if ((sp_pos = title.indexOf(' ')) == -1) return;
        
        scriptView.insertText(title.substring(0,sp_pos));
    }
    
    public void logwindow(Object sender)
    {
        LogBrowserController lbc = LogBrowserController.sharedLogBrowser();
        lbc.showWindow(null);
    }
	
	public void logdownload(Object sender)
	{
		LogDownloadController.sharedController().showWindow(null);
	}
    
    public void insertTextToScriptView(String str)
    {
		scriptView.insertText(str);
    }
    
    // deligate
    public void tableViewSelectionDidChange(NSNotification notification)
    {
        // [U[`lXg̍ڂIƌĂ΂܂B
        // notification͖܂B
        
        // IĂ鍀ڂȂlogin/out{^disableB
        // o^̃`lłIĂlogin{^enableBłȂdisableB
        // o^̃`Flłlogout{^enableBłȂdisableB
        
        // ڂȏ゠SelectAll{^enableB
        channel_selectall.setEnabled(channelListView.numberOfRows() > 0);
		/*System.out.println("getting num of rows");
		int num_of_channels_tmp = channelListView.numberOfRows();
		System.out.println("num of rows:"+num_of_channels_tmp);
		channel_selectall.setEnabled(num_of_channels_tmp > 0);
		System.out.println("set enabled of selectall to "+(num_of_channels_tmp > 0));*/
        
        textDidChange(null);
        
        if (channelListView.numberOfSelectedRows() == 0 || !Foundation.sharedFoundation().slpp_connected())
        {
            channel_login.setEnabled(false);
            channel_logout.setEnabled(false);
            return;
        }
        
        boolean regd_ch_found = false;
        boolean unregd_ch_found = false;
        
        int n_ch = channelListView.numberOfRows();
        
        Vector chList = Foundation.sharedFoundation().getSlppConnection().getChannelList();
        for (int i = 0;i < n_ch;i++)
        {
            if (channelListView.isRowSelected(i))
            {
                ChannelInfo info = (ChannelInfo)chList.elementAt(i);
                
                if (info.registered())
                {
                    regd_ch_found = true;
                }
                else
                {
                    unregd_ch_found = true;
                }
                
                if (regd_ch_found && unregd_ch_found)
                {
                    break; // ȏ̃[v͖ʁB
                }
            }
        }
        
        channel_logout.setEnabled(regd_ch_found);
        channel_login.setEnabled(unregd_ch_found);
    }
    
    public void textDidChange(NSNotification aNotification)
    {
        // SakuraScript\`FbNsAFB
        // notification͖܂B
        
        String data = scriptView.string();
        int len = data.length();
        boolean error = ScriptAnalyzer.analyze(scriptView);
        
        // ڑĂĕ͂Ăč\G[ăOC̃`lIĂԂȂ
        // Broadcast{^enableBIĂ`lNoPost炻̌ł͂ȂB
        // }`|Xg͋֎~ƂB
        if (data.length() > 0 && !error && channelListView.numberOfSelectedRows() == 1 && Foundation.sharedFoundation().slpp_connected())
        {
            int selected = channelListView.selectedRow();
            Vector chList = Foundation.sharedFoundation().getSlppConnection().getChannelList();
            ChannelInfo info = (ChannelInfo)chList.elementAt(selected);
            if (info.registered() && !info.nopost())
            {
                scriptsend_broadcast.setEnabled(true);
            }
            else
            {
                scriptsend_broadcast.setEnabled(false);
            }
        }
        else
        {
            scriptsend_broadcast.setEnabled(false);
        }
        
        // ͂Ăč\G[āeŏIĂLocalTest{^enableB
        scriptsend_localtest.setEnabled(data.length() > 0 && !error);
        
        // gXNvgłȂClear{^enableB
        //scriptsend_clear.setEnabled(!data.equals("\u00a5t\u00a5h\u00a5s[0]"));
		// ߁BenabledB
		
		// TCYZo
		try {
			int length_of_script = data.getBytes("SJIS").length;
			lengthofscript_current.setStringValue(length_of_script == 1 ? "1 byte" : length_of_script+" bytes");
			
			if (length_of_script > LIMIT_OF_SCRIPT_LENGTH) {
				scriptsend_broadcast.setEnabled(false);
			}
		} catch (Exception e) {}
    }
}
