/*
	$Id: LogData.java,v 1.1.1.1 2003/04/08 14:55:47 admin Exp $
	-----
	
    SSTP Bottle Client for MacOS X, using protocol SLPP2.0
    Copyright (C) 2001-2002 phonohawk.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package log;
import log.*;
import com.apple.cocoa.foundation.*;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.net.*;
import org.w3c.dom.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

/*
	bottlelogDTD͎̒ʂB
	<!ELEMENT bottlelog (message*)>
	<!ATTLIST bottlelog version CDATA #REQUIRED> <!-- 1.0 -->
	<!ATTLIST bottlelog generator CDATA #REQUIRED>
	<!ATTLIST bottlelog saved CDATA #REQUIRED> <!-- YYYY/MM/DD HH:MM:SS -->
	<!ELEMENT message (date, channel, script, votes, agrees, ghost)>
	<!ATTLIST message mid CDATA #REQUIRED>
	<!ELEMENT date (#PCDATA)>
	<!ELEMENT channel (#PCDATA)>
	<!ELEMENT script (#PCDATA)>
	<!ELEMENT votes (#PCDATA)>
	<!ELEMENT agrees (#PCDATA)>
	<!ELEMENT ghost (#PCDATA)>
	
	bottlelog-extraDTD͎̒ʂB
	<!ELEMENT bottlelog-extra (messagebyuser|messagebyadmin)*>
	<!ATTLIST bottlelog-extra version CDATA #REQUIRED> <!-- 1.0 -->
	<!ATTLIST bottlelog-extra generator CDATA #REQUIRED>
	<!ATTLIST bottlelog-extra saved CDATA #REQUIRED> <!-- YYYY/MM/DD HH:MM:SS -->
	<!ELEMENT messagebyuser (date,time,channel,script,ghost?)>
	<!ELEMENT messagebyadmin (date,time,script?,dialogmessage?)>
	<!ATTLIST messagebyadmin mid CDATA #REQUIRED>
	<!ELEMENT date (#PCDATA)> <!-- YYYYMMDD -->
	<!ELEMENT time (#PCDATA)> <!-- HHMMSS -->
	<!ELEMENT channel (#PCDATA)>
	<!ELEMENT script (#PCDATA)>
	<!ELEMENT ghost (#PCDATA)>
	<!ELEMENT dialogmessage (#PCDATA)>
*/

public class LogData
{
	public static final int LOG_ALL = 1;
	public static final int LOG_USER = 2;
	public static final int LOG_ADMIN = 3;
	
	public static final String GROUP_ALL = "All Bottles";
	public static final String GROUP_USER = "User Broadcasted";
	public static final String GROUP_ADMIN = "Administrator";
	
	public static final String[] DAYS_OF_WEEK =
		new String[] {"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
	
	int year;
	int month;
	int day;
	boolean for_all_is_dirty;
	boolean for_extra_is_dirty;
    Vector all; // content : (Hashtable)R[h
	Vector user; // content: (Hashtable)R[h
	Vector admin; // content: (Hashtable)R[h
	Vector groups; // <Object[]> , [0]:(String)name [1]:(Vector)bottles
    
    public LogData(int year,int month,int day)
    {
		this.year = year;
		this.month = month;
		this.day = day;
		this.for_all_is_dirty = false;
		this.for_extra_is_dirty = false;
		this.all = new Vector();
		this.user = new Vector();
		this.admin = new Vector();
		this.groups = new Vector();
		
		try {
			boolean xml_exists = false;
			File[] xml = getXmlFiles();
			
			if (xml[0].exists()) {
				xml_exists = true;
				try {
					loadXmlLogAll(new FileInputStream(xml[0]));
				}
				catch (Exception e) {
					System.err.println("Error in loading "+xml[0].getAbsolutePath());
					e.printStackTrace();
				}
			}
			else if (xml[1].exists()) {
				xml_exists = true;
				try {
					loadXmlLogAll(new GZIPInputStream(new FileInputStream(xml[1])));
				}
				catch (Exception e) {
					System.err.println("Error in loading "+xml[1].getAbsolutePath());
					e.printStackTrace();
				}

			}
			
			if (xml[2].exists()) {
				xml_exists = true;
				try {
					loadXmlLogExtra(new FileInputStream(xml[2]));
				}
				catch (Exception e) {
					System.err.println("Error in loading "+xml[2].getAbsolutePath());
					e.printStackTrace();
				}
			}
			else if (xml[3].exists()) {
				xml_exists = true;
				try {
					loadXmlLogExtra(new GZIPInputStream(new FileInputStream(xml[3])));
				}
				catch (Exception e) {
					System.err.println("Error in loading "+xml[3].getAbsolutePath());
					e.printStackTrace();
				}
			}
			
			if (!xml_exists) {
				File raw = getRawFile();
				File gzipped = getGzippedFile();
				if (raw.exists()) {
					loadFlatLog(new FileInputStream(raw));
				}
				else if (gzipped.exists()) {
					loadFlatLog(new GZIPInputStream(new FileInputStream(gzipped)));
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		
		reconstructGroups();
    }
	
	public void reconstructGroups()
	{
		synchronized (groups) {
			groups.clear();
			
			if (all.size() > 0) {
				groups.add(new Object[] {GROUP_ALL,all});
			}
			
			if (user.size() > 0) {
				groups.add(new Object[] {GROUP_USER,user});
			}
			
			if (admin.size() > 0) {
				groups.add(new Object[] {GROUP_ADMIN,admin});
			}
		}
	}
	
	public Vector getGroups()
	{
		synchronized (groups) {
			return groups;
		}
	}
	
	public File getRawFile() {
		return new File(LogDataServer.home,
				"Library/BotCli/log/"+year+"/"+(month<=9?"0"+month:""+month)+"/"+(day<=9?"0"+day:""+day)+".log");
	}
	
	public File getGzippedFile() {
		return new File(LogDataServer.home,
				"Library/BotCli/log/"+year+"/"+(month<=9?"0"+month:""+month)+"/"+(day<=9?"0"+day:""+day)+".log.gz");
	}
	
	public File[] getXmlFiles() {
		// [0]: <day>.xml
		// [1]: <day>.xml.gz
		// [2]: <day>-extra.xml
		// [3]: <day>-extra.xml.gz
		String base = "Library/BotCli/log/"+year+"/"+(month<=9?"0"+month:""+month)+"/"+(day<=9?"0"+day:""+day);
		File[] result = new File[4];
		result[0] = new File(LogDataServer.home,base+".xml");
		result[1] = new File(LogDataServer.home,base+".xml.gz");
		result[2] = new File(LogDataServer.home,base+"-extra.xml");
		result[3] = new File(LogDataServer.home,base+"-extra.xml.gz");
		return result;
	}
	
	public int getYear() {
		return year;
	}
	
	public int getMonth() {
		return month;
	}
	
	public int getDay() {
		return day;
	}
    
    protected void loadFlatLog(InputStream is)
    {
        // sǂݍHashtable쐬B
		DataInputStream dis = new DataInputStream(is);
        try
        {
			synchronized (all) {
				all.clear();
				while(true)
				{
					String line = dis.readUTF(); // EOFOďIB
					line = line.substring(0,line.length() - 1); // Ō̉sB
					
					Hashtable dict = new Hashtable();
				
					StringTokenizer st = new StringTokenizer(line,",");
					while(st.hasMoreTokens())
					{
						String token = st.nextToken();
						
						// label:data `ɂȂĂ̂ŁAȒOŕB
						String label = token.substring(0,token.indexOf(':'));
						String data = URLDecoder.decode(token.substring(token.indexOf(':') + 1),"SJIS");
						
						dict.put(label,data);
					}
					
					String type = (String)dict.get("type");
					if (type.equals("OB")) {
						all.add(dict);
					}
					else if (type.equals("UB")) {
						user.add(dict);
					}
					else if (type.equals("FB")) {
						admin.add(dict);
					}
				}
			}
        }
		catch (EOFException e)
		{
			// ok
		}
		catch (Exception e)
		{
			System.err.println("Error in loading log stream...");
			e.printStackTrace();
		}
		
		try{dis.close();}catch(Exception e){}
    }
	
	protected void loadXmlLogAll(InputStream is) throws Exception
	{
		SAXParserFactory factory = SAXParserFactory.newInstance();
		SAXParser parser = factory.newSAXParser();
		parser.parse(is,new Handler_bottlelog(all));
		is.close();
	}
	
	protected class Handler_bottlelog extends DefaultHandler
	{
		Vector bottles;
		Hashtable current_bottle;
		String current_tag;
		StringBuffer current_text;
		
		public Handler_bottlelog(Vector bottles)
		{
			this.bottles = bottles;
			this.current_bottle = null;
			this.current_text = new StringBuffer();
		}
		
		public void startElement(String uri,String localName,String qName,Attributes attributes)
		{
			if (qName.equals("message")) {
				// VȕrB
				current_bottle = new Hashtable();
				
				// midۑƓɁA͂datetimeB
				String mid = attributes.getValue("mid");
				current_bottle.put("mid",mid);
				
				Calendar cal = MidAnalyzer.analyzeTimeStampFromMid(mid);
				
				String date = pad(cal.get(Calendar.MONTH)+1,2)+pad(cal.get(Calendar.DAY_OF_MONTH),2)+pad(cal.get(Calendar.YEAR),4);
				current_bottle.put("date",date);
				
				String time = pad(cal.get(Calendar.HOUR_OF_DAY),2)+pad(cal.get(Calendar.MINUTE),2)+pad(cal.get(Calendar.SECOND),2);
				current_bottle.put("time",time);
			}
			else {
				current_tag = qName;
				current_text.setLength(0);
			}
		}
		
		public void characters(char[] ch,int offset,int length)
		{
			String content = new String(ch,offset,length).trim();
			current_text.append(content);
		}
		
		public void endElement(String uri,String localName,String qName)
		{
			if (qName.equals("message")) {
				// ̕rIB
				bottles.add(current_bottle);
			}
			else {
				String content = current_text.toString();
				
				if (qName.equals("channel")) {
					current_bottle.put("channel",content);
				}
				else if (current_tag.equals("script")) {
					current_bottle.put("script",content);
				}
				else if (current_tag.equals("votes")) {
					current_bottle.put("vote",content);
				}
				else if (current_tag.equals("agrees")) {
					current_bottle.put("agree",content);
				}
				else if (current_tag.equals("ghost")) {
					current_bottle.put("ifghost",content);
				}
			}
		}
	}
	
	protected void loadXmlLogExtra(InputStream is) throws Exception
	{
		SAXParserFactory factory = SAXParserFactory.newInstance();
		SAXParser parser = factory.newSAXParser();
		
		parser.parse(is,new Handler_bottlelog_extra(user,admin));
		is.close();
	}
	
	protected class Handler_bottlelog_extra extends DefaultHandler
	{
		Vector user,admin;
		Hashtable current_bottle;
		String current_tag;
		StringBuffer current_text;
		
		public Handler_bottlelog_extra(Vector user,Vector admin)
		{
			this.user = user;
			this.admin = admin;
			this.current_bottle = null;
			this.current_text = new StringBuffer();
		}
		
		public void startElement(String uri,String localName,String qName,Attributes attributes)
		{
			if (qName.equals("messagebyuser") || qName.equals("messagebyadmin")) {
				// VȕrB
				current_bottle = new Hashtable();
			}
			else {
				current_tag = qName;
				current_text.setLength(0);
			}
		}
		
		public void characters(char[] ch,int offset,int length)
		{
			String content = new String(ch,offset,length).trim();
			current_text.append(content);
		}
		
		public void endElement(String uri,String localName,String qName)
		{
			if (qName.equals("messagebyuser")) {
				// ̕rIB
				user.add(current_bottle);
			}
			else if (qName.equals("messagebyadmin")) {
				// ̕rIB
				admin.add(current_bottle);
			}
			else {
				String content = current_text.toString();
			
				if (current_tag.equals("date")) {
					current_bottle.put("date",content);
				}
				else if (current_tag.equals("time")) {
					current_bottle.put("time",content);
				}
				else if (current_tag.equals("channel")) {
					current_bottle.put("channel",content);
				}
				else if (current_tag.equals("script")) {
					current_bottle.put("script",content);
				}
				else if (current_tag.equals("ghost")) {
					current_bottle.put("ifghost",content);
				}
				else if (current_tag.equals("dialogmessage")) {
					current_bottle.put("fbmsg",content);
				}
			}
		}
	}
    
   	public void commit()
	{
		// ܂all
		if (for_all_is_dirty) {
			getRawFile().delete();
			getGzippedFile().delete();
			File[] xml = getXmlFiles();
			xml[0].delete();
			xml[1].delete();
			
			// g󂾂ۑȂB
			boolean empty = false;
			synchronized (all) {
				if (all.size() == 0) {
					empty = true;
				}
			}
			
			if (!empty) {
				try {
					if (LogFileCompressor.haveToCompress(year,month,day)) {
						// k
						xml[1].getParentFile().mkdirs();
						saveXmlLogAll(new GZIPOutputStream(new FileOutputStream(xml[1])));
						
					}
					else {
						// kȂB
						xml[0].getParentFile().mkdirs();
						saveXmlLogAll(new FileOutputStream(xml[0]));
					}
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
			
			for_all_is_dirty = false;
		}
		
		// extra
		if (for_extra_is_dirty) {
			getRawFile().delete();
			getGzippedFile().delete();
			File[] xml = getXmlFiles();
			xml[2].delete();
			xml[3].delete();
		
			// g󂾂ۑȂB
			boolean empty = false;
			synchronized (user) {
				synchronized (admin) {
					if (user.size() == 0 && admin.size() == 0) {
						empty = true;
					}
				}
			}
			
			if (!empty) {
				try {
					if (LogFileCompressor.haveToCompress(year,month,day)) {
						// k
						xml[3].getParentFile().mkdirs();
						saveXmlLogExtra(new GZIPOutputStream(new FileOutputStream(xml[3])));
					}
					else {
						// kȂB
						xml[2].getParentFile().mkdirs();
						saveXmlLogExtra(new FileOutputStream(xml[2]));
					}
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
			
			for_extra_is_dirty = false;
		}
		
		//System.out.println("commited: "+toString());
		System.gc();
	}
    
    protected void saveFlatLog(OutputStream os) throws Exception
    {
		// type:[^Cv],script:[XNvg],...[LF]
		// ^Cv"temporary-"Ŏn܂ĂۑȂB
		synchronized (all) {
			DataOutputStream dos = new DataOutputStream(os);
			final int n_records = all.size();
			for (int i = 0;i < n_records;i++)
			{
				Hashtable record = (Hashtable)all.elementAt(i);
				Object[] entries = record.entrySet().toArray(); // entries : <Map.Entry>
				StringBuffer buf = new StringBuffer();
				for (int j = 0;j < entries.length;j++)
				{
					Map.Entry entry = (Map.Entry)entries[j];
					
					String key = (String)entry.getKey();
					if (key.startsWith("temporary-")) continue;
					
					String value = (String)entry.getValue();
					
					buf.append((buf.length() > 0 ? "," : "")+key+":"+URLEncoder.encode(value,"SJIS"));
				}
				dos.writeUTF(buf.toString()+'\n');
			}
			dos.flush();
			dos.close();
		}
    }
	
	protected void saveXmlLogAll(OutputStream os) throws Exception
	{
		// oׂ͂̈DOMc[͔̂mȂA
		// ɗǂ@vtȂB
		// ͂œf̂̓GeBeBƂʓ|c
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		Document doc = builder.newDocument();
		
		String app_name = NSBundle.mainBundle().localizedStringForKey("CFBundleShortVersionString","BotCli","InfoPlist");
		
		Calendar cal = Calendar.getInstance();
		String saved = pad(cal.get(Calendar.YEAR),4)+'/'+pad(cal.get(Calendar.MONTH)+1,2)+'/'+pad(cal.get(Calendar.DAY_OF_MONTH),2)+' '+pad(cal.get(Calendar.HOUR_OF_DAY),2)+':'+pad(cal.get(Calendar.MINUTE),2)+':'+pad(cal.get(Calendar.SECOND),2);
		
		Element root = doc.createElement("bottlelog");
		root.setAttribute("version","1.0");
		root.setAttribute("generator",app_name);
		root.setAttribute("saved",saved);
		doc.appendChild(root);
		
		synchronized (all) {
			final int n_records = all.size();
			for (int i = 0; i < n_records; i++) {
				Hashtable record = (Hashtable)all.elementAt(i);
				
				Element message = doc.createElement("message");
				message.setAttribute("mid",emptyIfNull((String)record.get("mid")));
				root.appendChild(message);
				
				String date_str = (String)record.get("date");
				String time_str = (String)record.get("time");
				int month = Integer.parseInt(date_str.substring(0,2));
				int day = Integer.parseInt(date_str.substring(2,4));
				int year = Integer.parseInt(date_str.substring(4));
				int hour = Integer.parseInt(time_str.substring(0,2));
				int minute = Integer.parseInt(time_str.substring(2,4));
				int second = Integer.parseInt(time_str.substring(4));
				cal.set(Calendar.YEAR,year);
				cal.set(Calendar.MONTH,month-1);
				cal.set(Calendar.DAY_OF_MONTH,day);
				cal.set(Calendar.HOUR_OF_DAY,hour);
				cal.set(Calendar.MINUTE,minute);
				cal.set(Calendar.SECOND,second);
				
				String date_element = date_str.substring(4)+'/'+date_str.substring(0,2)+'/'+date_str.substring(2,4)+' '+time_str.substring(0,2)+':'+time_str.substring(2,4)+':'+time_str.substring(4)+" ("+DAYS_OF_WEEK[cal.get(Calendar.DAY_OF_WEEK)-1]+")";
				
				Element date = doc.createElement("date");
				date.appendChild(doc.createTextNode(date_element));
				message.appendChild(date);
				
				Element channel = doc.createElement("channel");
				channel.appendChild(doc.createTextNode(emptyIfNull((String)record.get("channel"))));
				message.appendChild(channel);
				
				Element script = doc.createElement("script");
				script.appendChild(doc.createTextNode(emptyIfNull((String)record.get("script"))));
				message.appendChild(script);
				
				Element votes = doc.createElement("votes");
				votes.appendChild(doc.createTextNode(zeroIfNull((String)record.get("vote"))));
				message.appendChild(votes);
				
				Element agrees = doc.createElement("agrees");
				agrees.appendChild(doc.createTextNode(zeroIfNull((String)record.get("agree"))));
				message.appendChild(agrees);
				
				Element ghost = doc.createElement("ghost");
				ghost.appendChild(doc.createTextNode(emptyIfNull((String)record.get("ifghost"))));
				message.appendChild(ghost);
			}
		}
		
		Transformer tr = TransformerFactory.newInstance().newTransformer();
		tr.setOutputProperty(OutputKeys.ENCODING,"UTF-8");
		tr.setOutputProperty(OutputKeys.INDENT,"yes");
		
		OutputStreamWriter osw = new OutputStreamWriter(os,"UTF-8");
		tr.transform(new DOMSource(doc),new StreamResult(osw));
		osw.flush();
		osw.close();
	}
	
	protected void saveXmlLogExtra(OutputStream os) throws Exception
	{
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		Document doc = builder.newDocument();
		
		String app_name = NSBundle.mainBundle().localizedStringForKey("CFBundleShortVersionString","BotCli","InfoPlist");
		
		Calendar cal = Calendar.getInstance();
		String saved = pad(cal.get(Calendar.YEAR),4)+'/'+pad(cal.get(Calendar.MONTH)+1,2)+'/'+pad(cal.get(Calendar.DAY_OF_MONTH),2)+' '+pad(cal.get(Calendar.HOUR_OF_DAY),2)+':'+pad(cal.get(Calendar.MINUTE),2)+':'+pad(cal.get(Calendar.SECOND),2);
		
		Element root = doc.createElement("bottlelog-extra");
		root.setAttribute("version","1.0");
		root.setAttribute("generator",app_name);
		root.setAttribute("saved",saved);
		doc.appendChild(root);
		
		// user
		synchronized (user) {
			final int n_records = user.size();
			for (int i = 0; i < n_records; i++) {
				Hashtable record = (Hashtable)user.elementAt(i);
				
				Element messagebyuser = doc.createElement("messagebyuser");
				root.appendChild(messagebyuser);
				
				Element date = doc.createElement("date");
				date.appendChild(doc.createTextNode(emptyIfNull((String)record.get("date"))));
				messagebyuser.appendChild(date);
				
				Element time = doc.createElement("time");
				time.appendChild(doc.createTextNode(emptyIfNull((String)record.get("time"))));
				messagebyuser.appendChild(time);
				
				Element channel = doc.createElement("channel");
				channel.appendChild(doc.createTextNode(emptyIfNull((String)record.get("channel"))));
				messagebyuser.appendChild(channel);
				
				Element script = doc.createElement("script");
				script.appendChild(doc.createTextNode(emptyIfNull((String)record.get("script"))));
				messagebyuser.appendChild(script);
				
				if (record.get("ifghost") != null) {
					Element ghost = doc.createElement("ghost");
					ghost.appendChild(doc.createTextNode((String)record.get("ifghost")));
					messagebyuser.appendChild(ghost);
				}
			}
		}
		
		// admin
		synchronized (admin) {
			final int n_records = admin.size();
			for (int i = 0; i < n_records; i++) {
				Hashtable record = (Hashtable)admin.elementAt(i);
				
				Element messagebyadmin = doc.createElement("messagebyadmin");
				root.appendChild(messagebyadmin);
				
				Element date = doc.createElement("date");
				date.appendChild(doc.createTextNode(emptyIfNull((String)record.get("date"))));
				messagebyadmin.appendChild(date);
				
				Element time = doc.createElement("time");
				time.appendChild(doc.createTextNode(emptyIfNull((String)record.get("time"))));
				messagebyadmin.appendChild(time);
				
				if (record.get("script") != null) {
					Element script = doc.createElement("script");
					script.appendChild(doc.createTextNode((String)record.get("script")));
					messagebyadmin.appendChild(script);
				}
				
				if (record.get("fbmsg") != null) {
					Element dialogmessage = doc.createElement("dialogmessage");
					dialogmessage.appendChild(doc.createTextNode(emptyIfNull((String)record.get("fbmsg"))));
					messagebyadmin.appendChild(dialogmessage);
				}
			}
		}
		
		Transformer tr = TransformerFactory.newInstance().newTransformer();
		tr.setOutputProperty(OutputKeys.ENCODING,"UTF-8");
		tr.setOutputProperty(OutputKeys.INDENT,"yes");
		
		OutputStreamWriter osw = new OutputStreamWriter(os,"UTF-8");
		tr.transform(new DOMSource(doc),new StreamResult(osw));
		osw.flush();
		osw.close();
	}
	
	public static String zeroIfNull(String str)
	{
		return (str != null ? str : "0");
	}
	
	public static String emptyIfNull(String str)
	{
		return (str != null ? str : "");
	}
	
	public static String pad(int n,int chars)
	{
		StringBuffer buf = new StringBuffer(chars);
		String str_n = Integer.toString(n);
		
		int n_pad = chars - str_n.length();
		for (int i = 0; i < n_pad; i++) {
			buf.append('0');
		}
		buf.append(str_n);
		
		return buf.toString();
	}
	
	public void add(int type,Hashtable bottle)
	{
		if (type == LOG_ALL) {
			synchronized (all) {
				all.add(bottle);
			}
			for_all_is_dirty = true;
		}
		else if (type == LOG_USER) {
			synchronized (user) {
				user.add(bottle);
			}
			for_extra_is_dirty = true;
		}
		else if (type == LOG_ADMIN) {
			synchronized (admin) {
				admin.add(bottle);
			}
			for_extra_is_dirty = true;
		}
		else {
			try {
				throw new Exception("Unexpected type: "+type);
			}
			catch (Exception e) {}
		}
		
		reconstructGroups();
	}
	
	public void beDirty(int type)
	{
		if (type == LOG_ALL) {
			for_all_is_dirty = true;
		}
		else if (type == LOG_USER || type == LOG_ADMIN) {
			for_extra_is_dirty = true;
		}
		else {
			try {
				throw new Exception("Unexpected type: "+type);
			}
			catch (Exception e) {}
		}
	}
    
    public Vector getArray()
	{
		return getArray(LOG_ALL);
	}
	public Vector getArray(int type)
    {
		// ԂꂽVectorMȂAbeDirty()Ă΂Ȃ΂ȂȂB
		// elements(),elementAt()ŃbN̓ǂׂBYޏłB
		if (type == LOG_ALL) {
			return all;
		}
		else if (type == LOG_USER) {
			return user;
		}
		else if (type == LOG_ADMIN) {
			return admin;
		}
		else {
			try {
				throw new Exception("Unexpected type: "+type);
			}
			catch (Exception e) {}
			return null;
		}
    }
	
	public Enumeration elements()
	{
		return elements(LOG_ALL);
	}
	public Enumeration elements(int type)
	{
		// ԂꂽHashtableMȂAbeDirty()Ă΂Ȃ΂ȂȂB
		if (type == LOG_ALL) {
			return all.elements();
		}
		else if (type == LOG_USER) {
			return user.elements();
		}
		else if (type == LOG_ADMIN) {
			return admin.elements();
		}
		else {
			try {
				throw new Exception("Unexpected type: "+type);
			}
			catch (Exception e) {}
			return null;
		}
	}
	
	public Hashtable elementAt(int i)
	{
		return elementAt(LOG_ALL,i);
	}
	public Hashtable elementAt(int type,int i)
	{
		// ԂꂽHashtableMȂAbeDirty()Ă΂Ȃ΂ȂȂB
		if (type == LOG_ALL) {
			return (Hashtable)all.elementAt(i);
		}
		else if (type == LOG_USER) {
			return (Hashtable)user.elementAt(i);
		}
		else if (type == LOG_ADMIN) {
			return (Hashtable)admin.elementAt(i);
		}
		else {
			try {
				throw new Exception("Unexpected type: "+type);
			}
			catch (Exception e) {}
			return null;
		}
	}
	
	public int size()
	{
		return size(LOG_ALL);
	}
	public int size(int type)
	{
		if (type == LOG_ALL) {
			return all.size();
		}
		else if (type == LOG_USER) {
			return user.size();
		}
		else if (type == LOG_ADMIN) {
			return admin.size();
		}
		else {
			try {
				throw new Exception("Unexpected type: "+type);
			}
			catch (Exception e) {}
			return 0;
		}
	}
	
	public String toString()
	{
		return year+"/"+month+"/"+day;
	}
}
